#include "web_config.h"

namespace config {
  iotwebconf::TextTParameter<STRING_LEN> stringParam = iotwebconf::Builder<iotwebconf::TextTParameter< STRING_LEN>>(
      "stringParam").label("String param").build();
}

//TODO: Is there any conflict with SPIFFS/LittleFS?
//TODO: Actually use those parameters
//TODO: Make sure values are *really* set to defaults during first run.

namespace web_config {
#if defined(ESP8266)
  ESP8266WebServer http(80); // Create a webserver object that listens for HTTP request on port 80
#elif defined(ESP32)
  WebServer http(80);
#endif

  DNSServer dnsServer; //TODO: Check if needed

  IotWebConf *iotWebConf;

// -- Configuration specific key. The value should be modified if config structure was changed.
  const char config_version[] = "ampel_test_v6";

  /**
   * Services
   */
  iotwebconf::ParameterGroup serviceParams = iotwebconf::ParameterGroup("Services", "Services");

  iotwebconf::CheckboxTParameter ampelWifiParam = iotwebconf::Builder<iotwebconf::CheckboxTParameter>("WiFi").label(
      "WiFi").defaultValue(true).build();
  iotwebconf::CheckboxTParameter ampelCSVParam =
      iotwebconf::Builder<iotwebconf::CheckboxTParameter>("CSV").label("CSV").defaultValue(true).build();
  iotwebconf::CheckboxTParameter ampelMQTTParam = iotwebconf::Builder<iotwebconf::CheckboxTParameter>("MQTT").label(
      "MQTT").defaultValue(true).build();
  iotwebconf::CheckboxTParameter ampelLoRaWANParam =
      iotwebconf::Builder<iotwebconf::CheckboxTParameter>("LoRaWAN").label("LoRaWAN").defaultValue(false).build();

  //TODO: Distribute to corresponding classes, possibly with callbacks?
  /**
   * CO2 sensor
   */
  iotwebconf::ParameterGroup co2Params = iotwebconf::ParameterGroup("co2", "CO2 Sensor"); //TODO: Conflict with labels?

  //TODO: UIntTParameter?
  iotwebconf::IntTParameter<uint16_t> timestepParam = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "timestep").label("Measurement timestep").defaultValue(MEASUREMENT_TIMESTEP).min(2).max(1800).step(1).placeholder(
      "2..1800").build();

  iotwebconf::FloatTParameter temperatureOffsetParam =
      iotwebconf::Builder<iotwebconf::FloatTParameter>("temp_offset").label("Temperature offset").defaultValue(
      TEMPERATURE_OFFSET).placeholder("-3").step(0.1).build();

  iotwebconf::IntTParameter<uint16_t> altitudeParam = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "altitude").label("Altitude").defaultValue(ALTITUDE_ABOVE_SEA_LEVEL).min(0).step(1).build();

  iotwebconf::IntTParameter<uint16_t> atmosphericCO2Param = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "atmospheric_co2").label("Atmospheric CO2 concentration").defaultValue(
  ATMOSPHERIC_CO2_CONCENTRATION).min(400).max(2000).step(1).placeholder("400..2000").build();

  iotwebconf::CheckboxTParameter autoCalibrateParam = iotwebconf::Builder<iotwebconf::CheckboxTParameter>("asc").label(
      "Auto-calibration").defaultValue(true).build();

  /**
   * CSV
   */
  iotwebconf::ParameterGroup csvParams = iotwebconf::ParameterGroup("csvs", "CSV");

  iotwebconf::IntTParameter<uint16_t> csvTimestepParam = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "csv_timestep").label("CSV timestep").defaultValue(CSV_INTERVAL).min(0).step(1).build();

  /**
   * MQTT
   */
  iotwebconf::ParameterGroup mqttParams = iotwebconf::ParameterGroup("mqtts", "MQTT");

  iotwebconf::IntTParameter<uint16_t> mqttTimestepParam = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "mqtt_timestep").label("MQTT timestep").defaultValue(
  MQTT_SENDING_INTERVAL).min(0).step(1).defaultValue(300).build();

  iotwebconf::CheckboxTParameter mqttCommandsParam = iotwebconf::Builder<iotwebconf::CheckboxTParameter>(
      "mqtt_commands").label("MQTT commands").defaultValue(false).build();

  iotwebconf::TextTParameter<STRING_LEN> mqttServerParam = iotwebconf::Builder<iotwebconf::TextTParameter<STRING_LEN>>(
      "mqtt_server").label("MQTT Server").defaultValue(MQTT_SERVER).build();

  iotwebconf::IntTParameter<uint16_t> mqttPortParam = iotwebconf::Builder<iotwebconf::IntTParameter<uint16_t>>(
      "mqtt_port").label("MQTT Port").defaultValue(MQTT_PORT).build();

  iotwebconf::TextTParameter<STRING_LEN> mqttUserParam = iotwebconf::Builder<iotwebconf::TextTParameter<STRING_LEN>>(
      "mqtt_user").label("MQTT User").defaultValue(MQTT_USER).build();

  iotwebconf::PasswordTParameter<STRING_LEN> mqttPasswordParam = iotwebconf::Builder<
      iotwebconf::PasswordTParameter< STRING_LEN>>("mqtt_password").label("MQTT password").defaultValue(
  MQTT_PASSWORD).build();

  /**
   * NTP Time
   */

  /**
   * LED
   */

  /**
   * LoRa & Stuff
   */

  void update() {
    iotWebConf->doLoop(); // Listen for HTTP requests from clients
  }

  void setWifiConnectionCallback(void (*function)()) {
    iotWebConf->setWifiConnectionCallback(function);
  }

  void setWifiConnectionFailedCallback(void (*function)()) {
    iotWebConf->setWifiConnectionFailedHandler([&function]() -> iotwebconf::WifiAuthInfo* {
      function();
      return NULL;
    });
  }

  void initialize() {
    iotWebConf = new IotWebConf(ampel.sensorId, &dnsServer, &http, HTTP_PASSWORD, config_version);

    serviceParams.addItem(&ampelWifiParam);
    serviceParams.addItem(&ampelMQTTParam);
    serviceParams.addItem(&ampelCSVParam);
    serviceParams.addItem(&ampelLoRaWANParam);

    co2Params.addItem(&timestepParam);
    co2Params.addItem(&temperatureOffsetParam);
    co2Params.addItem(&altitudeParam);
    co2Params.addItem(&atmosphericCO2Param);
    co2Params.addItem(&autoCalibrateParam);

    csvParams.addItem(&csvTimestepParam);

    mqttParams.addItem(&mqttTimestepParam);
    mqttParams.addItem(&mqttCommandsParam);
    mqttParams.addItem(&mqttServerParam);
    mqttParams.addItem(&mqttPortParam);
    mqttParams.addItem(&mqttUserParam);
    mqttParams.addItem(&mqttPasswordParam);

    iotWebConf->addParameterGroup(&serviceParams);
    iotWebConf->addParameterGroup(&co2Params);
    iotWebConf->addParameterGroup(&csvParams);
    iotWebConf->addParameterGroup(&mqttParams);

    sensor_console::defineCommand("reset_config", []() {
      Serial.println(F("Resetting config..."));
      //TODO: Reset every group
      iotWebConf->getSystemParameterGroup()->applyDefaultValue();
      iotWebConf->saveConfig();
      Serial.println(F("Done!"));
    }, F("(resets the complete IotWeb config)"));

    sensor_console::defineStringCommand("conf", [](char *new_value) {
      Serial.print(F("Setting stringParam to "));
      Serial.println(new_value);
      strncpy(config::stringParam.value(), new_value, STRING_LEN);
      Serial.println(F("Done"));
      iotWebConf->saveConfig();
    }, F("some_text (config setter)"));

#if !defined(AMPEL_WIFI)
    iotWebConf->loadConfig();
    return;
#endif

    const int ONBOARD_LED_PIN = 2;
# ifdef ESP8266
    iotWebConf->setStatusPin(ONBOARD_LED_PIN, LOW);
# else
    iotWebConf->setStatusPin(ONBOARD_LED_PIN, HIGH);
# endif
    iotWebConf->setWifiConnectionTimeoutMs(1000UL * WIFI_TIMEOUT);

#if defined(ESP8266)
  WiFi.hostname(ampel.sensorId);
#elif defined(ESP32)
    WiFi.setHostname(ampel.sensorId);
#endif

    iotWebConf->skipApStartup();
    //TODO: Add callbacks
    //TODO: Add LED effects
    //TODO: Allow offline config loading
    //TODO: Add default values for SSID/password
    //TODO: Add other params
    //TODO: Use HTTP_USER / HTTP_PASSWORD for config
    //TODO: Move to own class
    //TODO: Remove AP Password config?
    //TODO: Save LoRaWAN key if possible?
    //FIXME: Why does MQTT fail? (on ESP32)

    iotWebConf->init();

    //TODO: Authenticate only if required?
    //TODO: / captive fast return?

    http.on("/config", [] {
      iotWebConf->handleConfig();
    });

    http.onNotFound([]() {
      iotWebConf->handleNotFound();
    });
  }
}
